package Net::Stripe::Constants;
$Net::Stripe::Constants::VERSION = '0.43';
# ABSTRACT: Constants for use in validation and unit tests

use strict;

use constant MIN_API_VERSION  => '2011-01-01';
use constant MAX_API_VERSION  => '2020-03-02';

1;

__END__

=pod

=head1 NAME

Net::Stripe::Constants - Constants for use in validation and unit tests

=head1 VERSION

version 0.43

=head1 AUTHORS

=over 4

=item *

Luke Closs

=item *

Rusty Conover

=back

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2015 by Prime Radiant, Inc., (c) copyright 2014 Lucky Dinosaur LLC.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
